# Bogumił Kamiński, 2022-03-24
# Run under Julia 1.7.2

# vector u:
# u[1] = X
# u[2] = R
# u[3] = C(0,q-1), ..., u[3+q-1] = C(q-1,0)     -- functions of the form C(k1, k2) with k1+k2 = q-1
# u[3+q] = C(0,q), ..., u[3+q+q] = C(q,0)       -- functions of the form C(k1, k2) with k1+k2 = q

using DifferentialEquations # version: v7.1.0
using Plots # version 1.27.0

function L1_0(k1, k2, q, u)
    @assert k1+k2==q-1
    return k1 * u[2 + 1 + k1]
end

function L1_1(k1, k2, q, u)
    @assert k1+k2==q
    return k1 * u[2 + q + 1 + k1]
end

function L2_0(k1, k2, q, u)
    @assert k1+k2==q-1
    return k2 * u[2 + 1 + k1]
end

function L2_1(k1, k2, q, u)
    @assert k1+k2==q
    return k2 * u[2 + q + 1 + k1]
end

L1(u, q) = sum(k1 -> L1_0(k1, q-1-k1, q, u), 0:q-1) + sum(k1 -> L1_1(k1, q-k1, q, u), 0:q)
L2(u, q) = sum(k1 -> L2_0(k1, q-1-k1, q, u), 0:q-1) + sum(k1 -> L2_1(k1, q-k1, q, u), 0:q)

L3(u, q) = sum(k1 -> L1_0(k1, q-1-k1, q, u) * (q-1-k1), 0:q-1) +
           sum(k1 -> L1_1(k1, q-k1, q, u) * (q-k1), 0:q)

L4(u, q) = sum(k1 -> L2_0(k1, q-1-k1, q, u) * (q-1-k1), 0:q-1) +
           sum(k1 -> L2_1(k1, q-k1, q, u) * (q-k1), 0:q)

D(u, q)  = sum(k1 -> u[2 + 1 + k1], 0:q-1)

function our_ode!(du, u, q, t)
    @assert isodd(length(u))
    q = (length(u) - 3) ÷ 2

    # X'
    du[1] = 2 * (L1(u, q) + L2(u, q) + u[2]) +
            (1-u[1])
    # R'
    du[2] = -u[2] +
            L2(u, q) -
            2 * L1(u, q) * u[2] / (1-u[1]) +
            2 * L3(u, q) -
            2 * u[2] * ( 1 + u[2]/(1-u[1])) -
            2 * L2(u, q) * u[2] / (1-u[1]) +
#            2 * L2(u, q) * L2(u, q) / (1-u[1]) + # Algorithm 1
             2 * L4(u, q) +  # Algorithm 2
            2 * L2(u, q) * u[2] / (1-u[1]) -
            u[2]

    for k1 in 0:q-1
        k2 = q-1 - k1
        du[2 + 1 + k1] = -u[2 + 1 + k1] -
                         L2_0(k1,k2,q,u) +
                         (k1 > 0 && L2_0(k1-1,k2+1,q,u)) -
                         (u[1] - 5 * (u[2]+L1(u, q) + L2(u, q))) * u[2 + 1 + k1] / D(u, q) -
                         # Algortihm 1
#                         2 * L1_0(k1,k2, q, u) +
#                         (k1 > 0 && 2 * L1(u, q) * u[2 + k1] * (k2+1) / (1-u[1])) -
#                         2 * L1(u, q) * u[2 + 1 + k1] * k2 / (1-u[1]) -
#                         2 * (u[2]+L2(u, q)) * u[2 + 1 + k1] / (1-u[1]) +
#                         (k1 > 0 && 2 * L2_0(k1-1,k2+1, q, u)) -
#                         2 * L2_0(k1, k2, q, u) +
#                         (k1>0 && 2 * (u[2]+L2(u, q)) * u[2 + k1] * (k2+1) / (1-u[1])) -
#                         2 * (u[2]+L2(u, q)) * u[2 + 1 + k1] * k2 / (1-u[1]) -
                         # Algortihm 2
                         2 * ( L1_0(k1,k2, q, u) + L2_0(k1,k2, q, u) ) +
                         (k1 > 0 && 2 * ( L1(u, q) + L2(u, q) ) * u[2 + k1] * (k2+1) / (1-u[1])) -
                         2 * ( L1(u, q) + L2(u, q) ) * u[2 + 1 + k1] * k2 / (1-u[1]) -
                         2 * u[2] * u[2 + 1 + k1] / (1-u[1]) +
                         (k1>0 && 2 * u[2] * u[2 + k1] * (k2+1) / (1-u[1])) -
                         2 * u[2] * u[2 + 1 + k1] * k2 / (1-u[1]) -
                         # The rest
                         u[2] * u[2 + 1 + k1] / D(u, q) -
                         L1_0(k1, k2, q, u) +
                         (k2>0 && L1_0(k1+1,k2-1, q, u))
    end

    for k1 in 0:q
        k2 = q - k1
        du[2 + q + 1 + k1] = -u[2 + q + 1 + k1] -
                              L2_1(k1,k2, q, u) +
                              (k1 > 0 && L2_1(k1-1,k2+1, q, u)) +
                              (k1>0 && (u[1]-5 * (u[2]+L1(u, q)+L2(u, q))) * u[2 + k1] / D(u, q)) -
                              # Algorithm 1
#                              2 * L1_1(k1,k2,q,u) +
#                              (k1>0 && 2 * L1(u, q) * u[2 + q + k1] * (k2+1) / (1-u[1])) -
#                              2 * L1(u, q) * u[2 + q + 1 + k1] * k2 / (1-u[1]) -
#                              2 * (u[2]+L2(u, q)) * u[2 + q + 1 + k1] / (1-u[1]) +
#                              (k1>0 && 2 * L2_1(k1-1,k2+1, q, u)) -
#                              2 * L2_1(k1, k2, q, u) +
#                              (k1>0 && 2 * (u[2]+L2(u, q)) * u[2 + q + k1] * (k2+1) / (1-u[1])) -
#                              2 * (u[2]+L2(u, q)) * u[2 + q + 1 + k1] * k2 / (1-u[1]) +
                              # Algorithm 2
                              2 * ( L1_1(k1,k2,q,u) + L2_1(k1,k2,q,u) ) +
                              (k1>0 && 2 * (L1(u, q)+L2(u, q)) * u[2 + q + k1] * (k2+1) / (1-u[1])) -
                              2 * (L1(u, q)+L2(u, q)) * u[2 + q + 1 + k1] * k2 / (1-u[1]) -
                              2 * u[2] * u[2 + q + 1 + k1] / (1-u[1]) +
                              (k1>0 && 2 * u[2] * u[2 + q + k1] * (k2+1) / (1-u[1])) -
                              2 * u[2] * u[2 + q + 1 + k1] * k2 / (1-u[1]) +
                              # The rest
                              (k2>0 && u[2] * u[2 + 1 + k1] / D(u, q)) -
                              L1_1(k1, k2, q, u) +
                              (k2 > 0 && L1_1(k1+1,k2-1, q, u))
    end
end

function findlast_positive(x)
    x[1] < 0 && error("first value is negative")
    for i in eachindex(x::AbstractVector)
        x[i] < 0 && return i - 1
    end
    @warn "reached end of vector (a signal of potential numeric instability)"
    return length(x)
end

FLOAT_TYPE_USED = Float64 # use Float64 or BigFloat

res = []
t = []
res_l1 = []
res_l2 = []
u0 = FLOAT_TYPE_USED[0.0, # X
      0.0, # R
      1.0, # C(0,0)
      0.0, # C(1,0)
      0.0 # C(0,1)
     ];

q_ref = 1;
tspan0 = (FLOAT_TYPE_USED(0.0), FLOAT_TYPE_USED(1.6));
prob0 = ODEProblem(our_ode!, u0, tspan0);
sol0 = solve(prob0, dtmax=FLOAT_TYPE_USED(0.000001), maxiters=1e7);
max_i0 = minimum([findlast_positive(getindex.(sol0.u, i)) for i in 3:2+q_ref]);
@show D(sol0.u[max_i0], 1)
append!(res, sol0.u[1:max_i0]);
append!(t, sol0.t[1:max_i0]);
append!(res_l1, L1.(sol0.u[1:max_i0], 1))
append!(res_l2, L2.(sol0.u[1:max_i0], 1))
@show t[end], res[end][1], sum(res[end]) - res[end][2]

last_t = FLOAT_TYPE_USED(0.2)
for q_ref in 2:100
    @show q_ref
    u0 = zeros(FLOAT_TYPE_USED, 2 + q_ref + q_ref + 1);
    u0[1:2] = res[end][1:2];
    u0[3:2+q_ref] = res[end][end-q_ref+1:end];
    tspan0 = (0.0, last_t);
    prob0 = ODEProblem(our_ode!, u0, tspan0);
    sol0 = solve(prob0, dtmax=min(FLOAT_TYPE_USED(0.0000001), last_t/100), maxiters=1e7);
    max_i0 = minimum([findlast_positive(getindex.(sol0.u, i)) for i in 3:2+q_ref]);
    @show D(sol0.u[max_i0], q_ref)
    append!(res, sol0.u[1:max_i0]);
    global last_t = sol0.t[max_i0]
    append!(t, sol0.t[1:max_i0] .+ t[end]);
    append!(res_l1, L1.(sol0.u[1:max_i0], q_ref))
    append!(res_l2, L2.(sol0.u[1:max_i0], q_ref))
    @show t[end], res[end][1], sum(res[end]) - res[end][2]
end

# Cleanup phase

function final_ode!(du, u, q, t)
    du[1] = (u[3] - 5*(u[1] + u[2]) + 2*u[1] * (-1 - u[1]/(1 - u[3]) + 2*u[2]/(1 - u[3])) + 2*u[2]*(1 + 2*u[2]/(1 - u[3]) - u[1]/(1 - u[3])) - u[1]) + (2*u[2] - u[1])
    du[2] = (2*u[1])*(-2*u[2]/(1 - u[3])) + 2*u[2]*(-1 - 2*u[2]/(1 - u[3])) + u[1] - 2*u[2]
    du[3] = 1 - u[3] + 2*(u[1] + u[2])
end

q_final = q = (length(res[end]) - 3) ÷ 2
u0 = FLOAT_TYPE_USED[res[end][2] + L2(res[end], q_final), 0, res[end][1]]
tspan0 = (FLOAT_TYPE_USED(0.0), FLOAT_TYPE_USED(0.1)) # threshold is 0.0148849
prob0 = ODEProblem(final_ode!, u0, tspan0);
condition(u,t,integrator) = u[3] - 1
affect!(integrator) = terminate!(integrator)
cb = ContinuousCallback(condition,affect!)
sol0 = solve(prob0, dtmax=FLOAT_TYPE_USED(min(0.000001, tspan0[2]/100)), callback=cb)
t[end] + sol0.t[end]
